#include <SDL/SDL.h>
#define _REGION_ 1
#include "lvedes.h"
#include "region.h"

static region_t *rg_list_in[] = 
{
  NULL
};

//-- sorted copy of rg_list to get direct "index"-access --
static region_t** rg_list_sort = NULL;


region_t **rg_list_get()
{
  return rg_list_sort;
}


static T_VIDEO *video      = NULL;
static Uint8   *screen_dup = NULL;
static int     n_region    = 0;

void rg_list_init(T_VIDEO *pvideo)
{
  int		j, w, h;
  int           region_no;
  FILE		*fd;
  char 		fname[1024];
  SDL_Surface	*sf;
  Uint8		*p;
  region_t      *rg;
  int		n_rect = 0;

  //-- remind this ! -- 
  video = pvideo;

  w = video->screen->w;
  h = video->screen->h;
  
  //-- create (empty) region-screen same size as video->screen --
  if ((screen_dup = (Uint8 *)malloc(w * h)) == NULL)
  {
    fprintf(stderr, "Error: rg_list_init() - no memory for region screen\n");
    rg_list_cleanup();
    return;
  }
  memset(screen_dup, RG_ID_UNDEF, w * h);

  //-- count region elements --
  while (rg_list_in[n_region] != NULL)  n_region++;

  //-- create list for <n_region+1> sorted region elements --
  rg_list_sort = (region_t **)malloc((n_region+1) * sizeof(region_t *));  
  if (!rg_list_sort)
  {
    fprintf(stderr, "Error: rg_list_init() - no memory for sorted region list\n");
    rg_list_cleanup();
    return;  
  }
  rg_list_sort[0] = NULL;  // RG_ID_UNDEF
  
  //-- setup all region-elements (main-loop) ==
  //===========================================
  for (region_no = 0; region_no < n_region; region_no++) 
  {
    rg = rg_list_in[region_no];
  
    //== read coordinate definition(s) of region element ==
    //=====================================================     
    strcpy(fname, PATH_LIB);
    strcat(fname, rg->name);
    strcat(fname, RG_EXT_RECT);
    if ((fd = fopen(fname, "rb")) != NULL)
    {
      SDL_Rect *dst_rect;
    
      //== create array of rects definitions ==  
      //=======================================
      n_rect = rg->cols * rg->rows;
      dst_rect = (SDL_Rect *)malloc(n_rect * sizeof(SDL_Rect));
       
      //-- we have it ? -- 
      if (dst_rect != NULL)
      {    
        SDL_Rect b_rect;
        int      n_read, r, n, id;
        int      want_bmp;
         
        //-- bind this rect-array to region element -- 
        rg->dst_rect = dst_rect;
      
        //-- read rect defintions of first element (!) --
        n_read = fread(dst_rect, sizeof(SDL_Rect), 1, fd);

        //== region tables(!) needs extended definition ==
        //================================================
        if (n_read && (n_rect > 1))
        {
          #define COLS rg->cols
          #define ROWS rg->rows
        
          //== read table endpoint coordinats (of last rect) ==
          //===================================================
          if (fread(&b_rect, sizeof(SDL_Rect), 1, fd) == 1)
          {
            b_rect.w = b_rect.x - dst_rect[0].x;
            b_rect.h = b_rect.y - dst_rect[0].y;
 
            //-- calc brutto width and heights for one table element: --
            //-- b_rect.w = dst_rect.w + <gap_x>; 
            //-- b_rect.h = dst_rect.h + <gap_y>;  
            b_rect.w = dst_rect[0].w + (int)( (b_rect.w - dst_rect[0].w * COLS + 0.5)/(COLS-1.0) );
            b_rect.h = dst_rect[0].h + (int)( (b_rect.h - dst_rect[0].h * ROWS + 0.5)/(ROWS-1.0) );
            
            //== now calc coordinates of each table element starting at second ==
            //===================================================================  
            n = 1;
            for (r = 0; r < ROWS; r++)
            {  
              for (j = 0; j < COLS; j++)
              { 
                //-- first table element is already known --
                if (!j && !r) continue; 
                
                //-- calc coordinates by multiple of brutto width (height) --
                //-- in addition to the first-table element coordinates    --
                dst_rect[n].x = dst_rect[0].x + j * b_rect.w;
                dst_rect[n].y = dst_rect[0].y + r * b_rect.h;
                
                //-- same display width and height for each --
                dst_rect[n].w = dst_rect[0].w;
                dst_rect[n].h = dst_rect[0].h;
                
                n++;
              }    
            }  
          }
          else
          {
            fprintf(stderr, "Error: rg_list_init() - reading table data of (%s)\n", fname);
            
            //-- prevent unneeded allocation of region array --
            n_rect   = 1;             
            rg->cols = 1;
            rg->rows = 1;
          }
        }
        
        //== for this region fillup corresponend rects in <dup_screen> ==
        //===============================================================
        id = rg->id;
        for (j = 0; j < n_rect; j++)
        { 
          p  = screen_dup + (w * dst_rect[j].y) + dst_rect[j].x;
          for (r = 0; r < dst_rect[j].h; r++) 
          {
            memset(p, (Uint8)id, dst_rect[j].w);
            p += w;
          }
        }
	
        //== read bitmap file(s) if wanted ==
        //===================================        
        rg->sf[0] = NULL;
        rg->sf[1] = NULL;
        
        switch (rg->type)
        {
          case RG_TY_DRAG:
          case RG_TY_DROP:
            want_bmp = 1;
            break;
            
          default:
            want_bmp = 0;
            break;  
        }
        
        if (want_bmp)
        {
          strcpy(fname, PATH_LIB);
          strcat(fname, rg->name);
          strcat(fname, RG_EXT_BMP);
        
          rg->sf[0] = SDL_LoadBMP(fname);
          rg->sf[1] = SDL_LoadBMP(fname);
        
          sf = rg->sf[1];
          if (sf != NULL)
          {
            Uint8 bpp;
          
            rg->x = 0;
            rg->y = 0;
            rg->w = sf->w;
            rg->h = sf->h;
          
            bpp = sf->format->BytesPerPixel;
          
            p = (Uint8 *)sf->pixels;
            for (j = 0; j < (sf->w * sf->h * bpp); j++) p[j] = ~p[j];
          }
          else 
            fprintf(stderr, "Error: rg_list_init() - reading file (%s)\n", fname);
        } 
      }
      else
        fprintf(stderr, "Error: rg_list_init() - out of memory loading (%s)\n", fname);
       
      fclose(fd);  
    } 
    else
      fprintf(stderr, "Error: rg_list_init() - reading file %s\n", fname);
    
    //== for region tables we need to exchange the original region pointer ==
    //== with pointer of an region-array - each table-elemnt as a copy of  ==
    //== the original region but with its own <dst_rect> pointer selected  ==
    //== of the original regions rect array !                              ==
    //=======================================================================
    if (n_rect > 1)
    {
      region_t *rg_sav;
      
      rg_sav = rg;
      
      if ((rg = (region_t *)malloc(n_rect * sizeof(region_t))) != NULL)
      {
        for (j = 0; j < n_rect; j++)
        {
          memcpy(&(rg[j]), rg_sav, sizeof(region_t));
          rg[j].dst_rect = &(rg_sav->dst_rect[j]);
        }
      }
      else
        fprintf(stderr, "Error: rg_list_init() - out of memory creating (%s)-table\n", rg_sav->name);
    }
    
    //== store real region pointer in sorted list ==
    //============================================== 
    rg_list_sort[rg->id] = rg;
  
  } // for (region_no = ...) ends
}

void rg_list_cleanup()
{
  int i, n;

  i = 0;
  while (rg_list_in[i] != NULL)
  {
    for (n=0; n<2; n++)
    {
      if (rg_list_in[i]->sf[n])
        SDL_FreeSurface(rg_list_in[i]->sf[n]);
      rg_list_in[i]->sf[n] = NULL; 
    }
    
    if (rg_list_in[i]->dst_rect)  
      free(rg_list_in[i]->dst_rect);
    rg_list_in[i]->dst_rect = NULL;
    
    i++;
  }
  
  if (rg_list_sort) 
  {
    for (i = 0; i < n_region; i++)
    {  
      if (rg_list_sort[i]->rows > 1 || rg_list_sort[i]->cols > 1)
      {  
        free(rg_list_sort[i]);
        rg_list_sort[i] = NULL;
      }   
    }
    free(rg_list_sort);
    rg_list_sort = NULL;
  }
  
  if (screen_dup) free(screen_dup); 
  screen_dup = NULL;
}

region_t *rg_region_get(int id)
{
  return(rg_list_sort[id]);
}

void rg_region_draw(region_t *rg, int page)
{
  SDL_Surface   *dst_sf = video->screen;
  
  if ((page < 0) || (page > 1)) return;
  if (rg->page == page) return;
   
  rg->page = page;
 
  if (rg->dst_rect && rg->sf[page])
  {  
    SDL_BlitSurface(rg->sf[page], NULL, dst_sf, rg->dst_rect); 
    
    /* function is very buggy !
    SDL_UpdateRect(dst_sf, 
                 rg->dst_rect->x,
                 rg->dst_rect->y,
                 rg->dst_rect->x + rg->dst_rect->w,
                 rg->dst_rect->y + rg->dst_rect->h);
    */
    
    // alternativ function works better - but hangup program after a while 
    //SDL_UpdateRects(dst_sf, 1, rg->dst_rect);
    
    // this seems to work in every case (little bit oversized) --
    //SDL_UpdateRect(dst_sf, 0, 0, 0, 0);

    // all of the above updates will conflict with timer-callback
    // so we have to do the update in the callback routine itself
    video->refresh_bg = 1;
  }
}
                 
static void rg_region_invert(region_t *rg, int flag)
{
  if (rg->flag == flag) return;
  rg->flag = flag;
 
  if (rg->dst_rect)
  {
    Uint8 *p;  
    int   bpp, w;
    int   c, r, i;
    
    //-- get bytes per pixel --
    bpp = video->screen->format->BytesPerPixel;
    
    //-- get start offset in screen pixel-buffer --
    //-- and real wide (num of bytes)            --
    p = (Uint8 *)video->screen->pixels + rg->dst_rect->y * video->screen->pitch;
    p += (rg->dst_rect->x * bpp);
    w = rg->dst_rect->w * bpp;
    
    //-- invert rectangle inside the screen --
    for (r = 0; r < rg->dst_rect->h; r++) 
    {
      for (c = 0; c < w; c++)
      {
        p[c] = ~p[c];
      }
      p += video->screen->pitch; 
    }
    
    //-- we need refresh of display --
    video->refresh_bg = 1;
  }
}                 
 
region_t *rg_region_fit(int x, int y)
{
  int      id;
  region_t *rg;
  int      n = 0;
 
  id = screen_dup[video->screen->w * y + x];
  if ((rg = rg_region_get(id)) != NULL)
  {
    #define RG  rg[n]
    #define REC RG.dst_rect
       
    for (n = 0; n < (RG.cols * RG.rows); n++)     
    { 
      if ((x > REC->x) && (y > REC->y) && 
          (x < (REC->x + REC->w)) &&
          (y < (REC->y + REC->h)) )
        break;
    }
  }
  
  return(&(rg[n]));  
}

static region_t *rg_focus = NULL;
static void rg_single_focus(region_t * rg)
{
  if (rg_focus && (rg_focus != rg))
    rg_region_invert(rg_focus, RG_OFF);
  rg_focus = rg;
  
  rg_region_invert(rg, RG_ON);  
}

static region_t *rg_down = NULL;
static region_t *rg_drag  = NULL;
void rg_region_down(region_t *rg)
{
  rg_down = NULL;
  rg_drag = NULL;
 
  if (!rg) return;

  switch (rg->type)
  {
    case RG_TY_TOGGLE:
      rg_region_invert(rg, !rg->flag);
      break;
        
    case RG_TY_DOWN:
      rg_region_invert(rg, RG_ON);
      rg_down = rg;
      break;
      
    case RG_TY_DRAG:
      rg_single_focus(rg);
      rg_drag = rg;  
      break;
        
    case RG_TY_DROP:
      rg_single_focus(rg);    
      break;
        
    default:
      break;  
  }
}  

void rg_region_up(int x, int y)
{
  region_t *rg = rg_down;
  
  if (rg_drag)
  {
    if ((x > -1) && (y > -1))
    {  
      rg = rg_region_fit(x, y);
      if (rg && (rg->type != RG_TY_DROP))
      {
        rg = NULL;
      }  
    }
  }
    
  if (rg)  
  {
    switch (rg->type)
    {
      case RG_TY_DOWN:
        rg_region_invert(rg_down, RG_OFF);
        break;
      
      case RG_TY_DRAG:
        break;
        
      case RG_TY_DROP:
        rg_region_invert(rg_drag, RG_OFF);
        rg_single_focus(rg);
        break;
          
      default:
        //rg_region_invert(rg_down, RG_OFF);
        break;             
    }
  }
  
  rg_drag = NULL;
  rg_down = NULL;
}                          